/*
 Apache httpd module that uses RDMA to deliver data.

 This code is made available via the Apache License v2.0  (see LICENSE)
 and is Copyright 2007
		Dennis Dalessandro (dennis@osc.edu) - Ohio Supercomputer Center
		Pete Wyckoff (pw@osc.edu) - Ohio Supercomputer Center


 */

#include "ammasso.h"

int OK_TO_QUIT;
static const int RDMA_BLOCKING = 1;

void rnic_async_event_handler(cc_rnic_handle_t rnic_hndl, cc_event_record_t* event_record, void *user_context){
    info *my_info = (info*)user_context;

    cc_qp_query_attrs_t att;

    rnic_hndl = rnic_hndl;

    if((event_record->rnic_handle != my_info->rnic_hndl) || (event_record->rnic_user_context != user_context)){
	    /*Something went wonky and called this function when it should not have*/
	    am_error("Unknown RNIC handle and/or context passed to async event handler", CCERR_INVALID_RNIC);
    }

    switch (event_record->event_id) {
	    case CCAE_CONNECTION_REQUEST:
		    if((event_record->resource_indicator != CC_RES_IND_EP) ||
			    (event_record->resource_id.ep != my_info->listen_ep_hndl)){
			    am_error("Bad resource indicator in connection request", CCERR_INVALID_MODIFIER);
		    }
		    iwprintf(10,"[%s] Connect Request Received Details are:\n", __func__);
		    iwprintf(10,"\tladdr/lport %s/%d\n", inet_ntoa(*(struct in_addr*) &(event_record->event_data.connection_request.laddr)),
			    ntohs(event_record->event_data.connection_request.lport));
		    iwprintf(10,"\traddr/rport %s/%d\n", inet_ntoa(*(struct in_addr*) &(event_record->event_data.connection_request.raddr)),
			    ntohs(event_record->event_data.connection_request.rport));
		    iwprintf(10,"\tprivate_data=|%s|\n", event_record->event_data.connection_request.private_data);
		    my_info->cr_ep_hndl = event_record->event_data.connection_request.cr_handle;
		    my_info->state = CONNECTED;
		    break;
	    case CCAE_ACTIVE_CONNECT_RESULTS:
		    iwprintf(5,"[%s]Inside CCAE_CONNECT_RESULTS \n", __func__);

		    if(event_record->event_data.active_connect_results.status == CCERR_CONN_REFUSED){
			am_error("Connection refused", CCERR_CONN_REFUSED);
		    }

		    if(event_record->event_data.active_connect_results.status == CCERR_CONN_ABORTED){
			am_error("Connection aborted", CCERR_CONN_ABORTED);
		    }

		    if(event_record->resource_user_context != my_info){
			iwprintf(10,"WARNING - User context passed in does not match mine!\n");
		    }


		    cc_qp_query(my_info->rnic_hndl, my_info->qp_hndl, &att);



		    iwprintf(10, "Inside Async event handler qp state is %x which is ", att.qp_state);
		    if(att.qp_state == CC_QP_STATE_IDLE)
			iwprintf(10,"QP Idle\n");
		    else if(att.qp_state == CC_QP_STATE_CONNECTING)
			iwprintf(10,"QP Connecting\n");
		    else if(att.qp_state == CC_QP_STATE_RTS)
			iwprintf(10,"QP Ready to go\n");
		    else if(att.qp_state == CC_QP_STATE_CLOSING)
			iwprintf(10,"QP Closing\n");
		    else if(att.qp_state == CC_QP_STATE_TERMINATE)
			iwprintf(10,"QP is terminating\n");
		    else if(att.qp_state == CC_QP_STATE_ERROR)
			iwprintf(10,"QP is in an error state %x\n", CC_QP_STATE_ERROR);
		    else
			iwprintf(10,"Unknown QP state\n");




		    if(event_record->resource_indicator != CC_RES_IND_QP){
			iwprintf(0,"resource indicator is not a QP\n");
		    }

		    if(event_record->event_data.active_connect_results.status == CC_CONN_STATUS_SUCCESS){
			    iwprintf(10,"[%s] Active Connection Queried:\n", __func__);
			    iwprintf(10,"\tladdr/lport %s/%d\n", inet_ntoa(*(struct in_addr*)&(event_record->event_data.active_connect_results.laddr)),
				    ntohs(event_record->event_data.active_connect_results.lport));
			    iwprintf(10,"\traddr/rport %s/%d\n", inet_ntoa(*(struct in_addr*)&(event_record->event_data.active_connect_results.raddr)),
				    ntohs(event_record->event_data.active_connect_results.rport));
			    iwprintf(10,"\tprivate_data=|%s|\n", event_record->event_data.active_connect_results.private_data);
			    my_info->state = CONNECTED;
		    }


		    break;
	    case CCAE_LLP_CONNECTION_RESET:
		    am_error("Connection reset by otherside", CCERR_CONN_RESET);
		    break;
	    case CCAE_LLP_CLOSE_COMPLETE:
		    iwprintf(10,"[%s] Close complete, exiting",__func__);
		    OK_TO_QUIT = 1;
		    break;
	    default:
		    iwprintf(1,"[%s] Unexpected async event thrown:  \n", __func__);
		    exit(0);
		    am_error(cc_strevent(event_record->event_id), CC_OK);

		    cc_qp_query(my_info->rnic_hndl, my_info->qp_hndl, &att);
		    iwprintf(1, "TERMINATE MESSAGE IS %s\n", att.terminate_message);
		    //~ printf("send sgl depth is %d\n", att.send_sgl_depth);
		    //~ printf("recv sgl depth is %d\n", att.recv_sgl_depth);
    }


}

void am_error(const char *msg, cc_status_t code){

	fprintf(stderr,"----------ERROR----------\n");
	fprintf(stderr, "%s : %s\n", msg, cc_strerror(code));
	fprintf(stderr,"----------ERROR----------\n");
	fflush(stderr);

}

int open_rdma_dev(cc_rnic_handle_t *rnic_hndl, int flag, void *user_context){
    cc_status_t ret;

    info *rdma_info = user_context;

    ret = cc_rnic_open(0, CC_PBL_PAGE_MODE, rdma_info, rnic_hndl);
    if (ret != CC_OK) {
	    am_error("cc_rnic_open error", ret);
	    return -1;
    }


    if(user_context != NULL){
	rdma_info->rnic_open = OPEN;
    }

    if(flag == 1){
	ret = cc_eh_set_async_handler(*rnic_hndl, rnic_async_event_handler, rdma_info);
	if (ret != CC_OK) {
	    am_error("unable to set async event handler", ret);
	     return -1;
	}
    }
    iwprintf(5, "Rnic handle %x\n", *rnic_hndl);

    return 0;
}

int rdma_init(info *rdma_info){
    cc_qp_create_attrs_t attrs;
    int size;
    int ret;
    struct in_addr add_struct;
    cc_netaddr_t addr;
    cc_eh_ce_handler_id_t eh_id = 0;

    size = sizeof(addr);
    memset(&addr, 0, size);  /* avoid valgrind warnings later */

    ret = cc_rnic_getconfig(rdma_info->rnic_hndl, CC_GETCONFIG_ADDRS, &size, &addr);
    if(ret != CC_OK){
	am_error("Unable to get rnic config\n", ret);
	return -1;
    }

    rdma_info->host_type_addr = addr;

    if(RDMA_BLOCKING) {
	pthread_mutex_init(&(rdma_info->LOCK), NULL);
	pthread_mutex_init(&(rdma_info->READY), NULL);
	pthread_mutex_init(&(rdma_info->mutex_done), NULL);
	pthread_mutex_init(&(rdma_info->EVENT), NULL);
	pthread_cond_init(&(rdma_info->COND), NULL);
	ret = cc_eh_set_ce_handler(rdma_info->rnic_hndl, cqe_evh, &eh_id);
	if (ret != CC_OK){
	    am_error("can not create cqe handler", ret);
	    return -1;
	}
    }

    rdma_info->cq_depth = 20;
    if(RDMA_BLOCKING){
	ret = cc_cq_create(rdma_info->rnic_hndl, &rdma_info->cq_depth, eh_id, rdma_info, &rdma_info->cq_hndl);
	if (ret != CC_OK) {
		    am_error("can not create cq", ret);
		    return -1;
	}
    }
    else{
	ret = cc_cq_create(rdma_info->rnic_hndl, &rdma_info->cq_depth, 0, &rdma_info, &rdma_info->cq_hndl);
	if (ret != CC_OK) {
		    am_error("can not create cq", ret);
		    return -1;
	}

    }

    if(RDMA_BLOCKING) {
	/* request notification for next event */
	ret = cc_cq_request_notification(rdma_info->rnic_hndl, rdma_info->cq_hndl, CC_CQ_NOTIFICATION_TYPE_NEXT);
	if (ret != CC_OK){
	    am_error("cc_cq_request_notification", ret);
	    return -1;
	}
    }

    ret = cc_pd_alloc(rdma_info->rnic_hndl, &rdma_info->prot_domain_id);
    if (ret != CC_OK) {
	    am_error("cc_pd_alloc error ", ret);
	     return -1;
    }

    /*create the queue pair to use*/
    /*set qp attributes*/
    attrs.sq_cq = rdma_info->cq_hndl;
    attrs.rq_cq = rdma_info->cq_hndl;
    attrs.sq_depth = 250;
    attrs.rq_depth = 250;
    attrs.srq = 0;
    attrs.rdma_read_enabled = 1;
    attrs.rdma_write_enabled = 1;
    attrs.rdma_read_response_enabled = 1;
    attrs.mw_bind_enabled = 0;
    attrs.zero_stag_enabled = 0;
    attrs.send_sgl_depth = 1;
    attrs.recv_sgl_depth = 1;
    attrs.rdma_write_sgl_depth = 4;
    attrs.ord = 1;
    attrs.ird = 1;
    attrs.pdid = rdma_info->prot_domain_id;
    attrs.user_context = &rdma_info;

    ret = cc_qp_create(rdma_info->rnic_hndl, &attrs, &rdma_info->qp_hndl, &rdma_info->qp_id);
    if (ret != CC_OK) {
	am_error("cc_qp_create error", ret);
	return -1;
    }

    rdma_info->mapped = 0;
    rdma_info->rdma_bufsize = 0;
    rdma_info->outstanding_cqe = 0;

    add_struct.s_addr = addr.ip_addr;
    strcpy(rdma_info->ip_address, inet_ntoa(add_struct));
    iwprintf(5,"My IP is 0x%4x (%s)\n", ntohl(addr.ip_addr), rdma_info->ip_address);

    rdma_info->static_buffer = 0;  /*init this to 0 for now*/

    //~ iwprintf(5, "XXXXXXXXXXXXXXXXXXXXXxx cq hndl is %lld\n", rdma_info->cq_hndl);

    return 0;
}

int rdma_allocate_resources(info *rdma_info, int rdma_buf_size){

    int ret;

    if(__WORDSIZE != 64){
	iwprintf(0, "Unsupported on 32 bit platforms.\n");
	return -1;
    }
    if(rdma_buf_size != 0){
	rdma_info->rdma_buf = malloc(rdma_buf_size); /*there is our buffer*/
	iwprintf(5, "allocating buffer of %d bytes\n", rdma_buf_size);
	/*Register buffer to send and receive tagged messages (RDMA's)*/
	ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->rdma_buf, rdma_buf_size, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
	if (ret != CC_OK) {
		iwprintf(0, "Could not allocate %d bytes of memory\n", rdma_buf_size);
		am_error("cc_nsmr_register_virt error", ret);
		return -1;
	}

	rdma_info->rdma_to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;

	iwprintf(5,"my STag is %d\n", rdma_info->rdma_stag_index);
	iwprintf(5, "my TO is %lld\n", rdma_info->rdma_to);
	rdma_info->rdma_allocated = 1;
    }

    /*Get buffer ready to receive untagged message*/
    rdma_info->recv_buf = malloc(SIGNAL_SIZE);

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->recv_buf, SIGNAL_SIZE, rdma_info->prot_domain_id, 0, 0,
				    CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->recv_stag_index);
    if (ret != CC_OK) {
	    am_error("cc_nsmr_register_virt error", ret);
	    return -1;
    }

    rdma_info->recv_sgl.length = SIGNAL_SIZE;
    rdma_info->recv_sgl.stag = rdma_info->recv_stag_index;
    rdma_info->recv_sgl.to = (cc_uint64_t)(unsigned long)rdma_info->recv_buf;

    rdma_info->rq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rq_wr;
    rdma_info->rq_wr.local_sgl.sge_count = 1;
    rdma_info->rq_wr.local_sgl.sge_list = &rdma_info->recv_sgl;

    /*Get buffer ready to send untagged message*/
    rdma_info->send_buf = malloc(SIGNAL_SIZE);

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->send_buf, SIGNAL_SIZE, rdma_info->prot_domain_id, 0, 0,
				    CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->send_stag_index);
    if (ret != CC_OK) {
	    am_error("cc_nsmr_register_virt error", ret);
	    return -1;
    }

    rdma_info->send_sgl.length = SIGNAL_SIZE;
    rdma_info->send_sgl.stag = rdma_info->send_stag_index;
    rdma_info->send_sgl.to = (cc_uint64_t)(unsigned long)rdma_info->send_buf;

    rdma_info->sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->sq_wr;
    rdma_info->sq_wr.wr_type = CC_WR_TYPE_SEND;
    rdma_info->sq_wr.wr_u.send.local_sgl.sge_count = 1;
    rdma_info->sq_wr.wr_u.send.local_sgl.sge_list = &rdma_info->send_sgl;
    rdma_info->sq_wr.signaled = 1;

    memset(rdma_info->send_buf, '1', SIGNAL_SIZE);
    memset(rdma_info->recv_buf, '2', SIGNAL_SIZE);

    //~ iwprintf(5, "XXXXXXXXXXXXXXXXXXXXXxx cq hndl is %lld\n", rdma_info->cq_hndl);

    rdma_info->send_allocated =1 ;
    rdma_info->recv_allocated = 1;

    return 0;

}

int rdma_deallocate_resources(info *rdma_info){

    int ret;

    ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->rdma_stag_index);
    if(ret != CC_OK){
	am_error("cc_stag_dealloc error", ret);
	return -1;
    }

    ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->recv_stag_index);
    if(ret != CC_OK){
	am_error("cc_stag_dealloc error", ret);
	return -1;
    }

    ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->send_stag_index);
    if(ret != CC_OK){
	am_error("cc_stag_dealloc error", ret);
	return -1;
    }

    if(!rdma_info->mapped && rdma_info->rdma_allocated){
	free(rdma_info->rdma_buf);
	rdma_info->rdma_allocated = 0;
    }

    if(rdma_info->send_allocated){
	free(rdma_info->send_buf);
	rdma_info->send_allocated = 0;
    }

    if(rdma_info->recv_allocated){
	free(rdma_info->recv_buf);
	rdma_info->recv_allocated = 0;
    }

    return 0;

}

int rdma_connect(info *rdma_info, char *server, short unsigned int port){
    cc_qp_query_attrs_t att;
    unsigned char *private_data;
    int ret;
    in_addr_t addr;   /*inet style address for server*/
    short my_port;

    addr = inet_addr(server);
    my_port = htons(port);


    iwprintf(5,"connecting to %s on port %hu\n", server, port);

    /*prepare the private data*/
    private_data = malloc(7);
    memcpy(private_data, "server\0", 7);


    ret = cc_qp_connect(rdma_info->rnic_hndl, rdma_info->qp_hndl, addr , my_port, 7, private_data);
    if (ret != CC_OK) {
	    am_error("Unable to make connection request", ret);
	    return -1;
    }

    for(;;){
	/*attempt to connect the queue pair*/
	cc_qp_query(rdma_info->rnic_hndl, rdma_info->qp_hndl, &att);
	if(att.qp_state == CC_QP_STATE_RTS){
	    iwprintf(10,"QP Ready to go\n");
	    break;
	}
	else if(att.qp_state == CC_QP_STATE_CLOSING){
	    iwprintf(10,"QP Closing\n");
	    return -1;
	}
	else if(att.qp_state == CC_QP_STATE_TERMINATE){
	    iwprintf(10,"QP is terminating\n");
	    return -1;
	}
	else if(att.qp_state == CC_QP_STATE_ERROR){
	    iwprintf(10,"QP is in an error state %x\n", CC_QP_STATE_ERROR);
	    return -1;
	}
	//~ usleep(1);
    }
    rdma_info->state = CONNECTED;

    iwprintf(5,"Connection made, sent private data :%s:\n", private_data);

    free(private_data);
    return 0;
}

int rdma_listen_connection(info *rdma_info){
    unsigned short temp_port;
    int ret;
    in_addr_t addr;
    int addr_count;
    cc_netaddr_t local_addr;

    /*need a different way to get my host address*/
    addr_count = sizeof(local_addr);
    ret = cc_rnic_getconfig(rdma_info->rnic_hndl, CC_GETCONFIG_ADDRS, &addr_count, &local_addr);
    if(ret != CC_OK){
	am_error("unable to get rnic config for address", ret);
	exit(1);
    }

    addr = local_addr.ip_addr;

    /*start listening and let async event handler set state to connected*/
    rdma_info->state = IDLE;
    temp_port = 0;
    ret = cc_ep_listen_create(rdma_info->rnic_hndl, addr, &temp_port, CONN_BACK_LOG, rdma_info, &rdma_info->listen_ep_hndl);

    if(ret != CC_OK){
	am_error("Unable to listen on port", ret);
	return -1;
    }
    rdma_info->port = ntohs(temp_port);
    iwprintf(5,"Listening on port %d with back log of %d\n", rdma_info->port, CONN_BACK_LOG);


    return 0;

}

int rdma_wait_connection(info *rdma_info){
    int ret;
    unsigned char *private_data;
    iwprintf(5, "Waiting for connection....\n");
    while(1){
	if(rdma_info->state == CONNECTED)
	    break;
	iwprintf(10,"No one connected to RDMA device yet\n");
	//~ usleep(1);
    }

    iwprintf(5, "Connection requested\n");

    private_data = malloc(7);
    memcpy(private_data, "client\0", 7);
    ret = cc_cr_accept(rdma_info->rnic_hndl, rdma_info->cr_ep_hndl, rdma_info->qp_hndl, 7, private_data);
    if (ret != CC_OK) {
	    am_error("cc_cr_accept  - this was NOT expected", ret);
	    return -1;
    }

    iwprintf(5, "Connection made, sent private data :%s:\n", private_data);

    free(private_data);
    return 0;

}

int rdma_send_file(info *rdma_info, char *filename, int size){
    int fd;
    int ret;
    void *file_data;
    unsigned int posted;
    cc_wc_t wc;

    iwprintf(10,"sending %s which is %d bytes\n", filename, size);

    fd = open(filename, O_RDONLY);
    if(fd < 0){
	iwprintf(0,"can't open file\n");
	return -1;
    }

    file_data = mmap(0, size, PROT_READ, MAP_SHARED, fd, 0);
    if(file_data == (void *) -1){ /*mem map failed*/
	perror("Map of file failed with");
	return -1;
    }

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, file_data, size, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }

    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_sgl.stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sgl.to = (cc_uint64_t)(unsigned long)file_data;
    rdma_info->rdma_sgl.length = size;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_count = 1;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_list = &rdma_info->rdma_sgl;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_stag = rdma_info->remote_stag;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_to = rdma_info->remote_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.read_fence = 0;

    rdma_info->rdma_sq_wr.wr_type = CC_WR_TYPE_RDMA_WRITE;
    rdma_info->rdma_sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rdma_sq_wr;
    rdma_info->rdma_sq_wr.signaled = 1;

    iwprintf(5,"going to write to STag = %d and TO = %lld\n", rdma_info->remote_stag, rdma_info->remote_to);

    if(RDMA_BLOCKING){
	iwprintf(10,"requesting notification\n");
	ret = cc_cq_request_notification(rdma_info->rnic_hndl, rdma_info->cq_hndl,  CC_CQ_NOTIFICATION_TYPE_NEXT);
	if (ret != CC_OK){
	    am_error("cc_cq_request_notification", ret);
	    return -1;
	}
    }

    iwprintf(10,"posting send\n");
    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rdma_sq_wr, 1, &posted);
    if(ret != CC_OK || posted != 1){
	am_error("can't post work request", ret);
	return -1;
    }

    /*Poll the cq to tell me when its done*/
    iwprintf(10,"Polling CQ for RDMA Write to complete\n");
    ret = rdma_poll_cq(rdma_info, &wc, MAX_POLL, 0);
    if(ret != 0){
	iwprintf(10,"polling CQ failed miserably\n");
	return -1;
    }
    if(wc.wr_type != rdma_info->rdma_sq_wr.wr_type || wc.wr_id != rdma_info->rdma_sq_wr.wr_id){
	iwprintf(10,"Invalid work request returned in completion queue\n");
	return -1;
    }


    /*unmap the memory and get rid of STag*/
    ret = munmap(file_data, size);  /*remove the memory map*/
    if(ret != 0){
	iwprintf(10,"unable to unmap data\n");
	return -1;
    }

    close(fd);

    ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->rdma_stag_index);
    if(ret != CC_OK){
	iwprintf(10,"unable to deallocate stag\n");
	return -1;
    }

    iwprintf(10,"AHAHHHH EUREKA\n");

    return 0;

}

void cqe_evh(cc_rnic_handle_t rnic, cc_cq_handle_t cqh __attribute__((unused)), void *user_context){
	cc_status_t ret;
	info *my_info = user_context;
	/* need to request notification for getting rnic's response */
	rnic = rnic;
	iwprintf(10,"Dealing with completion queue %x here user context has %x\n", (unsigned) cqh, (unsigned)my_info->cq_hndl);

	ret = cc_cq_request_notification(my_info->rnic_hndl, my_info->cq_hndl,  CC_CQ_NOTIFICATION_TYPE_NEXT);
	if (ret != CC_OK)
		am_error("cc_cq_request_notification", ret);

		//~ iw_debug(3, "[%s] thread is not ready to poll the cq unlocking READY so thread can lock it", __func__);

	pthread_mutex_lock(&(my_info->LOCK));
	pthread_cond_signal(&(my_info->COND));
	pthread_mutex_unlock(&(my_info->LOCK));

}

int rdma_poll_cq(info *rdma_info, cc_wc_t *wc, u_int32_t polls, u_int32_t sleep_time) {
    cc_status_t cq_status;
    cc_rnic_handle_t rnic;
    uint32_t attempts = 0;
    cc_cq_handle_t cq;
    rnic = rdma_info->rnic_hndl;
    cq = rdma_info->cq_hndl;


    if(RDMA_BLOCKING){
	for(;;){
	    struct timespec ts;
	    int ret;
	    pthread_mutex_lock(&(rdma_info->LOCK));
	    cq_status = cc_cq_poll(rnic, cq, wc); /*check status*/
	    if(cq_status != CCERR_CQ_EMPTY) {
		pthread_mutex_unlock(&(rdma_info->LOCK));
		break;
	    }
	    iwprintf(5, "%s: blocking\n", __func__);
	    clock_gettime(CLOCK_REALTIME, &ts);
	    ts.tv_nsec += 50000000;  /* 50 ms */
	    if (ts.tv_nsec > 1000000000) {
		++ts.tv_sec;
		ts.tv_nsec -= 1000000000;
	    }
	    ret = pthread_cond_timedwait(&(rdma_info->COND),
	                                 &(rdma_info->LOCK), &ts);
	    if (ret)
		iwprintf(10, "%s: pthread_cond_timedwait: %m\n", __func__);
	    pthread_mutex_unlock(&(rdma_info->LOCK));
	}
	/*got something in the CQ lets see what it is*/
	if(cq_status != CC_OK){
		am_error("CQ status: Polling Completion Queue", cq_status);
		return -1;
	}
	if(wc->status != CC_OK){
		am_error("WC status: Polling Completion Queue", wc->status);
		return -1;
	}
	iwprintf(10,"Polling CQ Complete!!!!!!!!!!\n");

	return 0;
    }else {
	for(attempts = 0; polls == 0 || attempts <= polls; attempts++){
	    cq_status = cc_cq_poll(rnic, cq, wc); /*check status*/
	    if(cq_status != CCERR_CQ_EMPTY) /*something there*/
		    break;
	    if(sleep_time > 0)
		    usleep(sleep_time);
	}

	/*got something in the CQ lets see what it is*/
	if(cq_status != CC_OK){
	    if (cq_status == CCERR_CQ_EMPTY)
		am_error("Nothing ever showed up in CQ try increasing", cq_status);
	    else
		am_error("Error returned in completion queue event, cq_status != OK", cq_status);
	    return 1;
	}
	if(wc->status != CC_OK){
		am_error("Error returned in completion queue event, wc_status != OK", wc->status);
		return 1;
	}
	return 0;
    }
}

void iw_print_info(const char *fmt, ...){
    va_list ap;
    fprintf(stderr, "[iwarp_module]:");
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fflush(stderr);
    //~ fprintf(stderr, ".\n");
}

int send_rdma_buffer_info(int socket_fd, int rdma_buf_size, cc_stag_index_t stag, cc_uint64_t to){
    char r_buf[4096];
    int len;
    int bytes_sent;


    sprintf(r_buf, "\015\012HTTP/1.1 200 OK\015\012Content-Length: %d\015\012RDMA: stag=%d, to=%lld\015\012", rdma_buf_size, stag, to);
    strcat(r_buf, "\015\012");

    len = strlen(r_buf);
    bytes_sent = write(socket_fd, r_buf, len);

    if(bytes_sent != len)
	return -1;
    else
	return 0;



    return 0;
}

int rdma_map_file(info *rdma_info, char *filename, int size){
    int fd;
    int ret;
    unsigned int posted;

    iwprintf(10,"sending %s which is %d bytes\n", filename, size);

    fd = open(filename, O_RDONLY);
    if(fd < 0){
	iwprintf(0,"can't open file\n");
	return -1;
    }

    rdma_info->rdma_buf = mmap(0, size, PROT_READ, MAP_SHARED, fd, 0);
    if(rdma_info->rdma_buf == (void *) -1){ /*mem map failed*/
	perror("Map of file failed with");
	return -1;
    }

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->rdma_buf, size, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }

    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;

    iwprintf(5,"my STag is %d\n", rdma_info->rdma_stag_index);
    iwprintf(5, "my TO is %lld\n", rdma_info->rdma_to);

    /*before we send the buffer info, we need to post a receive so we can wait on it*/
    if(RDMA_BLOCKING){
	iwprintf(5,"requesting CQEH notification\n");
	ret = cc_cq_request_notification(rdma_info->rnic_hndl, rdma_info->cq_hndl,  CC_CQ_NOTIFICATION_TYPE_NEXT);
	if (ret != CC_OK){
	    am_error("cc_cq_request_notification", ret);
	    return -1;
	}
    }

    posted = 0;
    ret = cc_qp_post_rq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rq_wr, 1, &posted);
    if(ret != CC_OK) {
	am_error("cc_qp_post_rq", ret);
	return -1;
    }
    if(posted != 1) {
	printf("Error posting recieve didn't post 1 posted %d\n", ret);
	return -1;
    }




    return 0;

}

int rdma_read_file(info *rdma_info, int size){
    unsigned int posted;
    int ret;
    cc_wc_t wc;

    iwprintf(5, "Reading %d bytes from server\n", size);
    iwprintf(5, "STag = %d\n", rdma_info->remote_stag);
    iwprintf(5, "TO = %lld\n", rdma_info->remote_to);

    rdma_info->rdma_sgl.stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sgl.to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;
    rdma_info->rdma_sgl.length = size;

    rdma_info->rdma_sq_wr.wr_u.rdma_read.remote_stag = rdma_info->remote_stag;
    rdma_info->rdma_sq_wr.wr_u.rdma_read.remote_to = rdma_info->remote_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_read.local_stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sq_wr.wr_u.rdma_read.local_to = rdma_info->rdma_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_read.length = size;
    rdma_info->rdma_sq_wr.wr_type = CC_WR_TYPE_RDMA_READ;
    rdma_info->rdma_sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rdma_sq_wr;
    rdma_info->rdma_sq_wr.signaled = 1;

    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rdma_sq_wr, 1, &posted);
    if(ret != CC_OK || posted != 1){
	am_error("can't post work request\n", ret);
	return -1;
    }

    iwprintf(5, "polling for read completion\n");
    ret = rdma_poll_cq(rdma_info, &wc, 20*MAX_POLL, 0);
    if(ret != 0){
	am_error("polling completion queue\n", ret);
	return -1;
    }

    return 0;
}

int rdma_send_signal(info *rdma_info){

    int ret;
    cc_wc_t wc;
    unsigned int posted;

    iwprintf(5, "Sending signal..\n");
    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->sq_wr, 1, &posted);
    if (ret != CC_OK) {
	am_error("cc_qp_post_sq ", ret);
	return -1;
    }
    if (posted != 1) {
	iwprintf(0, "Trying to post send couldn't post 1, posted %d\n", posted);
	return -1;
    }

    iwprintf(5, "Signal is sent got to get its completion\n");


    ret = rdma_poll_cq(rdma_info, &wc, MAX_POLL, 0);
    if(ret != 0){
	iwprintf(0, "Can't poll CQ\n");
	return -1;
    }

    iwprintf(5, "Got CQE for the signal\n");

    return 0;
}

int rdma_wait_signal(info *rdma_info){
    cc_wc_t wc;
    int ret;

    iwprintf(5, "Waiting (polling CQ) for signal from client\n");

    ret = rdma_poll_cq(rdma_info, &wc, MAX_POLL, 0);
    if(ret != 0){
	iwprintf(0, "Can't poll CQ\n");
	return -1;
    }

    iwprintf(5, "Got signal from user guess I'm done\n");

    return 0;
}

int rdma_write_from_file(info *rdma_info, int fd, int offset, int len){
    //~ void *file_data;
    int ret;
    unsigned int posted;

    /*since we are using mmpa offset should be a multiple of page size if not warn, and continue anyway*/
    if(offset / getpagesize() != 0){
	printf("WARING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
    }

    rdma_info->rdma_buf = mmap(0, len, PROT_READ, MAP_SHARED, fd, offset);

    if(rdma_info->rdma_buf == (void *) -1){ /*mem map failed*/
	    perror("Map of file failed with");
	    return -1;
    }
    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->rdma_buf, len, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }
    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_sgl.stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sgl.to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;
    rdma_info->rdma_sgl.length = len;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_count = 1;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_list = &rdma_info->rdma_sgl;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_stag = rdma_info->remote_stag;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_to = rdma_info->remote_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.read_fence = 0;

    rdma_info->rdma_sq_wr.wr_type = CC_WR_TYPE_RDMA_WRITE;
    rdma_info->rdma_sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rdma_sq_wr;
    rdma_info->rdma_sq_wr.signaled = 1;

    iwprintf(5,"going to write to STag = %d and TO = %lld\n", rdma_info->remote_stag, rdma_info->remote_to);

    iwprintf(10,"posting send\n");
    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rdma_sq_wr, 1, &posted);
    if(ret != CC_OK || posted != 1){
	am_error("can't post work request", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;

    rdma_info->mapped = 1;
    rdma_info->rdma_bufsize = len;

    //~ /*unmap the memory and get rid of STag*/
    //~ ret = munmap(file_data, len);  /*remove the memory map*/
    //~ if(ret != 0){
	//~ iwprintf(10,"unable to unmap data\n");
	//~ return -1;
    //~ }


    //~ ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->rdma_stag_index);
    //~ if(ret != CC_OK){
	//~ iwprintf(10,"unable to deallocate stag\n");
	//~ return -1;
    //~ }

    iwprintf(10,"AHAHHHH EUREKA\n");
	return 0;

}

int rdma_write_from_buffer(info *rdma_info, void *buffer, int len){
    int ret;
    unsigned int posted;

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, (unsigned char *)buffer, len, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }

    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_sgl.stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sgl.to = (cc_uint64_t)(unsigned long)buffer;
    rdma_info->rdma_sgl.length = len;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_count = 1;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_list = &rdma_info->rdma_sgl;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_stag = rdma_info->remote_stag;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_to = rdma_info->remote_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.read_fence = 0;

    rdma_info->rdma_sq_wr.wr_type = CC_WR_TYPE_RDMA_WRITE;
    rdma_info->rdma_sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rdma_sq_wr;
    rdma_info->rdma_sq_wr.signaled = 1;

    iwprintf(5,"going to write to STag = %d and TO = %lld\n", rdma_info->remote_stag, rdma_info->remote_to);

    iwprintf(10,"posting send\n");
    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rdma_sq_wr, 1, &posted);
    if(ret != CC_OK || posted != 1){
	am_error("can't post work request", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;


    //~ /*unmap the memory and get rid of STag*/
    //~ ret = munmap(buffer, len);  /*remove the memory map*/
    //~ if(ret != 0){
	//~ iwprintf(10,"unable to unmap data\n");
	//~ return -1;
    //~ }


    //~ ret = cc_stag_dealloc(rdma_info->rnic_hndl, rdma_info->rdma_stag_index);
    //~ if(ret != CC_OK){
	//~ iwprintf(10,"unable to deallocate stag\n");
	//~ return -1;
    //~ }

    iwprintf(10,"AHAHHHH EUREKA\n");

    return 0;
}

int rdma_reap_completions(info *rdma_info){

    int ret = 0;
    cc_wc_t wc;
    iwprintf(5, "Getting %d completion event(s)\n", rdma_info->outstanding_cqe);
    while (rdma_info->outstanding_cqe) {
	ret = rdma_poll_cq(rdma_info, &wc, 0, 0);
	if (ret != 0) {
	    break;
	}
	--rdma_info->outstanding_cqe;
	iwprintf(5, "Got a completion event\n");
    }

    return ret;
}

int rdma_read_map_file(info *rdma_info, int fd, int offset, int size){

    int ret;
    unsigned int posted;

     if(offset / getpagesize() != 0){
	printf("WARING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
    }

    rdma_info->rdma_buf = mmap(0, size, PROT_READ, MAP_SHARED, fd, offset);
    if(rdma_info->rdma_buf == (void *) -1){ /*mem map failed*/
	perror("Map of file failed with");
	return -1;
    }

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, rdma_info->rdma_buf, size, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }

    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;

    iwprintf(5,"RDMA READ MAP FILE: my STag is %d\n", rdma_info->rdma_stag_index);
    iwprintf(5,"RDMA READ MAP FILE: my TO is %lld\n", rdma_info->rdma_to);

    /*before we send the buffer info, we need to post a receive so we can wait on it*/

    posted = 0;
    ret = cc_qp_post_rq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rq_wr, 1, &posted);
    if(ret != CC_OK) {
	am_error("cc_qp_post_rq", ret);
	return -1;
    }
    if(posted != 1) {
	printf("Error posting recieve didn't post 1 posted %d\n", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;
    rdma_info->mapped = 1;
    rdma_info->rdma_bufsize = size;

    return 0;
}

int rdma_read_map_buffer(info *rdma_info, void *buffer, int size){
    int ret;
    unsigned int posted;

    ret = cc_nsmr_register_virt(rdma_info->rnic_hndl, CC_ADDR_TYPE_VA_BASED, buffer, size, rdma_info->prot_domain_id, 0, 1,
				   CC_ACF_REMOTE_READ|CC_ACF_REMOTE_WRITE|CC_ACF_LOCAL_READ|CC_ACF_LOCAL_WRITE, &rdma_info->rdma_stag_index);
    if (ret != CC_OK) {
	am_error("cc_nsmr_register_virt error", ret);
	return -1;
    }

    iwprintf(10,"buffer is pinned and ready to go\n");

    rdma_info->rdma_to = (cc_uint64_t)(unsigned long)buffer;

    iwprintf(5,"my STag is %d\n", rdma_info->rdma_stag_index);
    iwprintf(5, "my TO is %lld\n", rdma_info->rdma_to);

    /*before we send the buffer info, we need to post a receive so we can wait on it*/

    posted = 0;
    ret = cc_qp_post_rq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rq_wr, 1, &posted);
    if(ret != CC_OK) {
	am_error("cc_qp_post_rq", ret);
	return -1;
    }
    if(posted != 1) {
	printf("Error posting recieve didn't post 1 posted %d\n", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;

    return 0;



    return 0;
}

int rdma_write_from_file_static(info *rdma_info, int fd, int offset, int len){
     int ret;
    unsigned int posted;
    void *file_buffer;
    void *ret_ptr;

    /*since we are using mmpa offset should be a multiple of page size if not warn, and continue anyway*/
    if(offset / getpagesize() != 0){
	printf("WARING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
    }

    iwprintf(5, "Buffer is already pinned no need to do anything other than copy the file into it\n");
    file_buffer = mmap(0, len, PROT_READ, MAP_SHARED, fd, offset);
    if(file_buffer == (void *) -1){ /*mem map failed*/
	    perror("Map of file failed with");
	    return -1;
    }

    /*copy the file into the rdma buffer*/
    ret_ptr = memcpy(rdma_info->rdma_buf, file_buffer, len);
     if(ret_ptr == (void *) -1){ /*mem map failed*/
	    perror("Map of file failed with");
	    return -1;
    }

    rdma_info->rdma_sgl.stag = rdma_info->rdma_stag_index;
    rdma_info->rdma_sgl.to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;
    rdma_info->rdma_sgl.length = len;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_count = 1;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.local_sgl.sge_list = &rdma_info->rdma_sgl;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_stag = rdma_info->remote_stag;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.remote_to = rdma_info->remote_to;
    rdma_info->rdma_sq_wr.wr_u.rdma_write.read_fence = 0;

    rdma_info->rdma_sq_wr.wr_type = CC_WR_TYPE_RDMA_WRITE;
    rdma_info->rdma_sq_wr.wr_id = (cc_uint64_t)(unsigned long)&rdma_info->rdma_sq_wr;
    rdma_info->rdma_sq_wr.signaled = 1;

    iwprintf(5,"going to write to STag = %d and TO = %lld\n", rdma_info->remote_stag, rdma_info->remote_to);

    iwprintf(10,"posting send\n");
    ret = cc_qp_post_sq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rdma_sq_wr, 1, &posted);
    if(ret != CC_OK || posted != 1){
	am_error("can't post work request", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;

    rdma_info->rdma_bufsize = len;



    return 0;
}

int rdma_read_map_file_static(info *rdma_info, int fd, int offset, int size){
    int ret;
    unsigned int posted;
    void *ret_ptr;
    void *file_buffer;

     if(offset / getpagesize() != 0){
	printf("WARING OFFSET IS NOT A MULTIPLE OF PAGE SIZE\n");
    }

    iwprintf(5, "Buffer is already pinned no need to do anything other than copy the file into it\n");
    file_buffer = mmap(0, size, PROT_READ, MAP_SHARED, fd, offset);
    if(file_buffer == (void *) -1){ /*mem map failed*/
	    perror("Map of file failed with");
	    return -1;
    }

    /*copy the file into the rdma buffer*/
    ret_ptr = memcpy(rdma_info->rdma_buf, file_buffer, size);
     if(ret_ptr == (void *) -1){ /*mem map failed*/
	    perror("Map of file failed with");
	    return -1;
    }

    rdma_info->rdma_to = (cc_uint64_t)(unsigned long)rdma_info->rdma_buf;

    iwprintf(5,"RDMA READ MAP FILE: my STag is %d\n", rdma_info->rdma_stag_index);
    iwprintf(5,"RDMA READ MAP FILE: my TO is %lld\n", rdma_info->rdma_to);

    /*before we send the buffer info, we need to post a receive so we can wait on it*/

    posted = 0;
    ret = cc_qp_post_rq(rdma_info->rnic_hndl, rdma_info->qp_hndl, &rdma_info->rq_wr, 1, &posted);
    if(ret != CC_OK) {
	am_error("cc_qp_post_rq", ret);
	return -1;
    }
    if(posted != 1) {
	printf("Error posting recieve didn't post 1 posted %d\n", ret);
	return -1;
    }

    rdma_info->outstanding_cqe++;
    rdma_info->rdma_bufsize = size;

    return 0;
}

int rdma_cleanup(info *rdma_info){
    cc_status_t ret;

    iwprintf(5, "cleaning up RDMA resources\n");

    ret = cc_qp_destroy(rdma_info->rnic_hndl, rdma_info->qp_hndl);
    if(ret != CC_OK){
	am_error("Can't destroy QP", ret);
    }

    ret = cc_pd_dealloc(rdma_info->rnic_hndl, rdma_info->prot_domain_id);
    if(ret != CC_OK){
	am_error("can't deallocate PD", ret);
    }

    ret = cc_cq_destroy(rdma_info->rnic_hndl, rdma_info->cq_hndl);
    if(ret != CC_OK){
	am_error("Can't destory CQ", ret);
    }


    iwprintf(5,"Done cleaning up rdma resources\n");
    return 0;
}

int rdma_zero_info(info *rdma_info){

    /*anything else that needs to be initialized for EACH connection do it in here*/
    rdma_info->not_tcp = 0;
    rdma_info->send_allocated = 0;
    rdma_info->recv_allocated = 0;
    rdma_info->rdma_allocated = 0;

    return 0;
}

